package weka.clusterers;

import weka.clusterers.*;
import weka.classifiers.libsvm.*;
import weka.core.*;
import java.util.*;

public class FarthestKernelSampler 
{       
    private FastVector m_sampled = null;
    private FastVector m_toSampleFrom = null;    
    
    public void start(Enumeration instEnum, Random rndm, Object initial) throws Exception 
    {
        this.m_sampled = new FastVector();
        this.m_toSampleFrom = new FastVector();
        // collect the instances
        while (instEnum.hasMoreElements()) {
            this.m_toSampleFrom.addElement((InstanceContainer)(instEnum.nextElement()));
        }
        // init the sampled
        if (initial != null) {
            // we take what the user gave
            FastVector initialSampled = (FastVector)initial;
            Enumeration initEnum = initialSampled.elements();
            while (initEnum.hasMoreElements()) {
                this.m_sampled.addElement((InstanceContainer)(initEnum.nextElement()));
            }            
        } else if (rndm != null) {
            // first one is chosen randomly
            int first = (int)Math.ceil((this.m_toSampleFrom.size()-1) * rndm.nextDouble());
            this.m_sampled.addElement((InstanceContainer)(this.m_toSampleFrom.elementAt(first)));
            this.m_toSampleFrom.removeElementAt(first);
        } else {
            // we just take the first one
            this.m_sampled.addElement((InstanceContainer)(this.m_toSampleFrom.elementAt(0)));
            this.m_toSampleFrom.removeElementAt(0);
        }        
    }
    
    public InstanceContainer next(svm_parameter kernelParams) throws Exception 
    {
        int i, farthestIndx = -1;
        int size = this.m_toSampleFrom.size();
        double maxDist = -Double.MAX_VALUE, currDist;
        InstanceContainer ins, farthestIns=null;
        
        for (i=0; i<size; i++) {
            ins = (InstanceContainer)(this.m_toSampleFrom.elementAt(i));
            currDist = this.distFromSet(ins, this.m_sampled, kernelParams);
            if (Utils.gr(currDist, maxDist)) {
                maxDist = currDist;
                farthestIndx = i;
            }            
        }
        farthestIns = (InstanceContainer)(this.m_toSampleFrom.elementAt(farthestIndx));
        this.m_toSampleFrom.removeElementAt(farthestIndx);
        this.m_sampled.addElement(farthestIns);
        return farthestIns;
    }
    
    public double[] nextDistribution(svm_parameter kernelParams, 
                                     DataContainer pool, 
                                     FastVector initSample) throws Exception 
    {
        return this.nextDistribution(kernelParams,
                                     pool,
                                     initSample,
                                     100);
    }
    
    public double[] nextDistribution(svm_parameter kernelParams, 
                                     DataContainer pool, 
                                     FastVector initSample,
                                     double scaling) throws Exception 
    {        
        int i;
        int size = pool.size();
        double currDist, sum;
        double lowestDist = Double.MAX_VALUE, highestDist=-Double.MAX_VALUE;
        InstanceContainer ins;
        double[] prob = new double[size];
        
        System.out.println("kfar dist");
        // find all instances confidence and scale by min-max      
        for (i=0; i<size; i++) {
            ins = pool.getInstance(i);
            currDist = this.distFromSet(ins, initSample, kernelParams);
            prob[i] = currDist;              
            if (Utils.sm(prob[i], lowestDist)) {
                lowestDist = prob[i];
            }
            if (Utils.gr(prob[i], highestDist)) {
                highestDist = prob[i];
            }            
        }                    
        for (i=0; i<size; i++) {
            if (Utils.eq(lowestDist, highestDist)) {
                prob[i] = 0.5;
            } else {
                prob[i] = (prob[i] - lowestDist) / (highestDist - lowestDist);
                if (Utils.sm(prob[i], 0.0)) {
                    prob[i] = 0.0;
                } 
                if (Utils.gr(prob[i], 1.0)) {
                    prob[i] = 1.0;
                }
            }                 
        }    
        // reverse the probs, so the lower becomes the higher                
        for (i=0; i<size; i++) {                  
            prob[i] = 1.0 - prob[i];              
        }  
        // scale up             
        double scale = scaling;        
        for (i=0; i<size; i++) {
            prob[i] *= scale;            
        }     
        // scale exponentially 
        sum = 0.0;        
        for (i=0; i<size; i++) {
            prob[i] = Math.exp(-1.0 * prob[i]);            
            sum += prob[i];            
        }               
        for (i=0; i<size; i++) {
            prob[i] = prob[i] / sum;               
        }
        for (i=0; i<size; i++) {
            if (Utils.grOrEq(prob[i], 0.01)) {                
                System.out.println("prob[" + i + "]=" + prob[i]);
            }            
        }        
        // check
        sum = 0.0;
        for (i=0; i<size; i++) {            
            sum += prob[i];
        } 
        if (!Utils.eq(sum, 1.0)) throw new Exception("sum probs is not 1.0 - sum=" + sum + "size=" + size);
        return prob;                        
    } 
    
    public void end() throws Exception 
    {
        this.m_sampled = null;
        this.m_toSampleFrom = null;
    }    
    
    private double distFromSet(InstanceContainer ins, 
                               FastVector instSet,
                               svm_parameter kernelParams) throws Exception
    {
        double minDist = Double.MAX_VALUE;
        double currDist;
        InstanceContainer setIns;
        Enumeration setEnum = instSet.elements();
        
        while (setEnum.hasMoreElements()) {
            setIns = (InstanceContainer)(setEnum.nextElement());
            currDist = this.distance(ins, setIns, kernelParams);
            //currDist = this.noKernelDistance(ins, setIns);
            if (Utils.sm(currDist, minDist)) {
                minDist = currDist;
            }
        }
        return minDist;
    }
    
    private double distance(InstanceContainer ins1, 
                            InstanceContainer ins2,
                            svm_parameter kernelParams) throws Exception
    {
        double dist;
        svm_node[] x = ((SvmLibPair)ins1).getNodes();
        svm_node[] y = ((SvmLibPair)ins2).getNodes();
        double kxx = svm.kernelEval(x, x, kernelParams, false);
        kxx += svm.g_maxFix;
        double kyy = svm.kernelEval(y, y, kernelParams, false);
        kyy += svm.g_maxFix;
        double kxy = svm.kernelEval(x, y, kernelParams, false);
        dist = kxx + kyy - 2*kxy;
        dist = Math.sqrt(dist);        
        return dist;
    }
    
    private double noKernelDistance(InstanceContainer ins1,
                                    InstanceContainer ins2) throws Exception
    {
        double dist;
        svm_node[] x = ((SvmLibPair)ins1).getNodes();
        svm_node[] y = ((SvmLibPair)ins2).getNodes();        
        dist = SvmLibPair.distance(x, y);
        return dist;
    }
}
